%%%
%%% Forward/inverse transforms with IntDCT based on direct-lifting
%%%                                                       by Taizo Suzuki
%%%
%%% Taizo Suzuki and Masaaki Ikehara,
%%% "Integer DCT based on direct-lifting of DCT-IDCT for
%%%                                       lossless-to-lossy image coding,"
%%% IEEE Trans. Image Process., Vol. 19, No. 11, pp. 2958-2965, Nov. 2010.
%%% DOI: 10.1109/TIP.2010.2051867
%%%

bs = 8; % block size
load BARBARA.mat % input image X

[H, W] = size(X);
s = zeros(bs); % side information block to be transmitted

% Forward Transform in Encoder
Xt = zeros(H, W); % transformed image
for h = 1 : bs : H-1
    for w = 1 : bs : W-1
        x = X(h:h+bs-1, w:w+bs-1);
        s = s + round( DCTf(x, bs));
        x = x + round(-DCTi(s, bs));
        s = s + round( DCTf(x, bs));
        tmp = s;
        s = -x;
        x = tmp;
        Xt(h:h+bs-1, w:w+bs-1) = x;
    end
end
Xt = block2sub(Xt, bs); % block mode to subband mode

% Arbitrary Lossless Encoding & Decoding
Yt = Xt; % if no encoding & decoding
t  = s;  % if no encoding & decoding

% Inverse Transform in Decoder
Yt = sub2block(Yt, bs); % subband mode to block mode
Y = zeros(H, W); % reconstructed image
for h = H-bs+1 : -bs : 1
    for w = W-bs+1 : -bs : 1
        y = Yt(h:h+bs-1, w:w+bs-1);
        tmp = t;
        t = y;
        y = -tmp;
        t = t - round( DCTf(y, bs));
        y = y - round(-DCTi(t, bs));
        t = t - round( DCTf(y, bs));
        Y(h:h+bs-1, w:w+bs-1) = y;
    end
end

dB = psnr(X, Y); % PSNR
fprintf('PSNR = %d [dB]\n', dB)

%% Forward DCT
function Y = DCTf(X, bs)

C = DCT2matrix(bs);
Y = C * X  * C.';

end

%% Inverse DCT
function Y = DCTi(X, bs)

C = DCT2matrix(bs).';
Y = C * X  * C.';

end

%% DCT-II matrix
function C = DCT2matrix(bs)

C = eye(bs);
for n = 0 : bs - 1
    for m = 0 : bs - 1
        beta_m = 1;
        if m == 0
            beta_m = 1 / sqrt(2);
        end
        C(m + 1, n + 1) = sqrt(2 / bs) * beta_m * cos(pi * m * (n + 1 / 2) / bs);
    end
end

end

%% Block Mode to Subband Mode
function Y = block2sub(X, bs)

[H, W] = size(X);
Y = [];
for h = 1 : bs
    Z = [];
    for w = 1 : bs
        YY = X(h:bs:H, w:bs:W);
        Z = [Z, YY];
    end
    Y = [Y; Z];
end

end

%% Subband Mode to Block Mode
function Y = sub2block(X, bs)

[H, W] = size(X);
Y = zeros(H, W);
HH = H / bs;
WW = W / bs;
for h = 1 : bs
    for w = 1 : bs
        Y(h:bs:H, w:bs:W) = X(HH*(h-1)+1:HH*h, WW*(w-1)+1:WW*w);
    end
end

end